from .base_path_processor import BasePathProcessor
from magnesium.utils import SnapshootableStack
from magnesium.node import Tag, Attribute
from magnesium.query_path import QueryPath


class SimplePathProcessor(BasePathProcessor):
    """"""
    
    def __init__(self, query_marker_strategy=None, text_prop_name='text'):
        """"""

        self.query_marker_strategy = query_marker_strategy
        self.text_prop_name = text_prop_name
    
    def process(self, element):
        """"""

        stack = SnapshootableStack()
        query_paths = {}

        self.process_element(stack, query_paths, element)

        return query_paths
    
    def process_element(
        self,
        stack: SnapshootableStack,
        query_paths: dict,
        element
    ):
        """"""

        tag_name = element.tag
        tag = Tag(tag_name)
        attributes = element.attrib
        text = element.text.strip()

        stack.push(tag)

        queried_attributes = []

        for k, v in attributes.items():
            attribute = Attribute(k, v)
            if self.query_marker_strategy.check(v):
                queried_attributes.append(attribute)
            else:
                tag.add_attribute(attribute)

        for attribute in queried_attributes:
            var_name = self.query_marker_strategy.get(attribute.value)
            snapshot = stack.snapshot()
            query_path = QueryPath(snapshot, attribute.name)
            query_paths[var_name] = query_path

        if self.query_marker_strategy.check(text):
            var_name = self.query_marker_strategy.get(text)
            snapshot = stack.snapshot()
            query_path = QueryPath(snapshot, self.text_prop_name)
            query_paths[var_name] = query_path

        for child in element:
            self.process_element(stack, query_paths, child)

        stack.pop()



class Pipeline(object):
    """"""

    def __init__(self, steps):
        """"""

        self.steps = steps
    
    def execute(self, data):
        """"""

        results = data

        for step in self.steps:
            results = step.execute(results)

        return results



from .base_path_interpreter import BasePathInterpreter


class XPathInterpreter(BasePathInterpreter):
    """"""

    def __init__(self, text_prop_name='text'):
        """"""

        self.text_prop_name = text_prop_name

    def interpret(self, expression):
        """"""

        path = ''
        
        for tag in expression.path:
            path += f'//{tag.name}'
            for attribute in tag.attributes:
                path += f'[@{attribute.name}="{attribute.value}"]'
        
        queried_prop = expression.queried_prop

        if queried_prop == self.text_prop_name:
            path += '/text()'
        else:
            path += f'/@{queried_prop}'

        return f'{path}'



from .attribute import Attribute


class Tag(object):
    def __init__(self, name):
        self.name = name
        self.attributes = []
        
    def add_attribute(self, attribute: Attribute):
        self.attributes.append(attribute)



from .base_pipeline_builder import BasePipelineBuilder
from .pipeline import Pipeline
from .pipeline_step import PipelineStep


class ExtractorPipelineBuilder(BasePipelineBuilder):
    """"""

    def __init__(self):
        """"""

        self.post_processors = []

    def set_mapping(self, v):
        """"""

        self.mapping = v
        return self

    def set_extractor(self, v):
        """"""

        self.extractor = v
        return self

    def add_post_processor(self, v, i=None):
        """"""

        if i is None:
            self.post_processors.append(v)
        else:
            self.post_processors.insert(i, v)
        return self

    def build(self):
        """"""

        self.extractor.mapping = self.mapping

        extractor_step = PipelineStep(
            self.extractor.extract
        )

        post_processor_steps = [
            PipelineStep(post_processor.process)
            for post_processor in self.post_processors
        ]

        steps = [
            extractor_step,
            *post_processor_steps
        ]

        return Pipeline(
            steps=steps
        )



from .base_query_marker_strategy import BaseQueryMarkerStrategy


class DoubleCurlyQueryMarkerStrategy(BaseQueryMarkerStrategy):
    """"""

    def check(self, x):
        """"""

        return x.startswith('{{') and x.endswith('}}')
    
    def get(self, x):
        """"""

        return x[2:-2].strip()



from .base_post_processor import BasePostProcessor
from .list_of_dicts_processor import ListOfDictsProcessor
from .pandas_data_frame_processor import PandasDataFrameProcessor
from .space_normalizer_processor import SpaceNormalizerProcessor
from .tail_cutter_processor import TailCutterProcessor



from lxml import etree

from .base_prefab import BasePrefab

from magnesium.query_marker import DoubleCurlyQueryMarkerStrategy
from magnesium.path_processor import SimplePathProcessor
from magnesium.path_interpreter import XPathInterpreter
from magnesium.mapping import SimpleMapping
from magnesium.pipeline import (
    SchemaPipelineBuilder,
    ExtractorPipelineBuilder
)
from magnesium.extractor import XPathExtractor
from magnesium.post_processor import (
    SpaceNormalizerProcessor,
    TailCutterProcessor,
    PandasDataFrameProcessor
)


class PandasPrefab(BasePrefab):
    """"""

    def __init__(
        self,
        schema_data=None,
        query_marker_strategy=DoubleCurlyQueryMarkerStrategy(),
        path_processor=SimplePathProcessor(),
        path_interpreter=XPathInterpreter(),
        mapping=SimpleMapping(),
        extractor_processor=XPathExtractor()
    ):
        """"""

        self.query_marker_strategy = query_marker_strategy
        self.path_processor = path_processor
        self.path_interpreter = path_interpreter
        self.mapping = mapping
        self.extractor_processor = extractor_processor

        self.set_schema_data(schema_data)
        
    def set_schema_data(self, v):
        self.schema_data = v

        schema_root = etree.fromstring(self.schema_data)

        schema_pipeline_builder = SchemaPipelineBuilder()
        schema_pipeline_builder.set_query_marker_strategy(
            self.query_marker_strategy
        ).set_text_prop_name(
            'text'
        ).set_path_processor(
            self.path_processor
        ).set_path_interpreter(
            self.path_interpreter
        ).set_mapping(
            self.mapping
        )

        schema_pipeline = schema_pipeline_builder.build()
        self.fit_mapping = schema_pipeline.execute(schema_root)

    def execute(self, sample_data):
        """"""

        sample_root = etree.fromstring(sample_data)

        space_normalizer_processor = SpaceNormalizerProcessor()
        tail_cutter_processor = TailCutterProcessor()
        pandas_data_frame_processor = PandasDataFrameProcessor()

        extractor_pipeline_builder = ExtractorPipelineBuilder()
        extractor_pipeline_builder.set_mapping(
            self.fit_mapping
        ).set_extractor(
            self.extractor_processor
        ).add_post_processor(
            tail_cutter_processor
        ).add_post_processor(
            space_normalizer_processor
        ).add_post_processor(
            pandas_data_frame_processor
        )

        extractor_pipeline = extractor_pipeline_builder.build()
        results = extractor_pipeline.execute(sample_root)

        return results



from .base_pipeline_builder import BasePipelineBuilder
from .extractor_pipeline_builder import ExtractorPipelineBuilder
from .pipeline_step import PipelineStep
from .pipeline import Pipeline
from .schema_pipeline_builder import SchemaPipelineBuilder



class SnapshootableStack(object):
    def __init__(self):
        self.stack = []
        
    def pop(self):
        return self.stack.pop()
    
    def push(self, item):
        self.stack.append(item)
    
    def snapshot(self):
        return self.stack.copy()



from .base_mapping import BaseMapping


class SimpleMapping(BaseMapping):
    """"""
    
    def __init__(self, path_interpreter=None):
        """"""

        self.path_interpreter = path_interpreter
        
    def create_map(self, query_paths):
        """"""

        results = {}
        for k, v in query_paths.items():
            results[k] = self.path_interpreter.interpret(v)
        return results



class PipelineStep(object):
    """"""

    def __init__(self, func):
        """"""

        self.func = func
    
    def execute(self, data):
        """"""

        results = self.func(data)
        return results



from abc import ABC, abstractmethod


class BaseQueryMarkerStrategy(ABC):
    """"""

    @abstractmethod
    def check(self, x):
        """"""

    @abstractmethod
    def get(self, x):
        """"""



from .base_pipeline_builder import BasePipelineBuilder
from .pipeline import Pipeline
from .pipeline_step import PipelineStep


class SchemaPipelineBuilder(BasePipelineBuilder):
    """"""

    def __init__(self):
        """"""

        self.post_processors = []

    def set_query_marker_strategy(self, v):
        """"""

        self.query_marker_strategy = v
        return self

    def set_text_prop_name(self, v):
        """"""

        self.text_prop_name = v
        return self

    def set_path_processor(self, v):
        """"""

        self.path_processor = v
        return self

    def set_path_interpreter(self, v):
        """"""

        self.path_interpreter = v
        return self

    def set_mapping(self, v):
        """"""

        self.mapping = v
        return self

    def build(self):
        """"""

        self.path_processor.query_marker_strategy = self.query_marker_strategy
        self.path_processor.text_prop_name = self.text_prop_name
        path_processor_step = PipelineStep(self.path_processor.process)

        self.path_interpreter.text_prop_name = self.text_prop_name
        self.mapping.path_interpreter = self.path_interpreter
        mapping_step = PipelineStep(self.mapping.create_map)

        steps = [
            path_processor_step,
            mapping_step,
        ]

        return Pipeline(steps=steps)



from .base_post_processor import BasePostProcessor


class ListOfDictsProcessor(BasePostProcessor):
    """"""

    def process(self, results):
        """"""

        processed_results = []

        keys = list(results.keys())
        values = list(results.values())

        for j in range(len(values[0])):
            d = {}
            for i in range(len(keys)):
                d[keys[i]] = values[i][j]
            processed_results.append(d)

        return processed_results



from .base_post_processor import BasePostProcessor


class TailCutterProcessor(BasePostProcessor):
    """"""

    def process(self, results):
        """"""

        processed_results = {}

        for k, v in results.items():
            processed_results[k] = [item for item in v if not item.is_tail]

        return processed_results



from .base_extractor import BaseExtractor


class XPathExtractor(BaseExtractor):
    """"""
    
    def __init__(self, mapping=None):
        """"""

        self.mapping = mapping

    def extract(self, data):
        """"""

        results = {}

        for k, v in self.mapping.items():
            results[k] = data.xpath(v)

        return results



from .base_post_processor import BasePostProcessor
import pandas as pd


class PandasDataFrameProcessor(BasePostProcessor):
    """"""
    
    def process(self, results):
        """"""
        
        processed_results = pd.DataFrame(results)        
        return processed_results
